#import "MAServiceControl.h"

#include <Security/Authorization.h>
#include <sys/wait.h>

#import "MAdministrator.h"
#include "mahelper.h"

static int
helperPerformCommand(AuthorizationRef authorizationRef, 
                     const char *helperPath, MAHelperCommand command)
{
  int tohelper[2], fromhelper[2];
  int childStatus = 0;
  int written;
  pid_t pid;
  AuthorizationExternalForm extAuth;
  
  // authref --> bytestream
  if (AuthorizationMakeExternalForm(authorizationRef, &extAuth))
    return MAHelperCommandInternalError;
  
  if (pipe(tohelper) < 0)
    return MAHelperCommandInternalError;
 
  if (pipe(fromhelper) < 0)
  {
    close(tohelper[0]);
    close(tohelper[1]);
    return MAHelperCommandInternalError;
  }
  
  if ((pid = fork()) < 0)
  {
    close(tohelper[0]);
    close(tohelper[1]);
    close(fromhelper[0]);
    close(fromhelper[1]);
    return MAHelperCommandInternalError;
  }
  else if (pid == 0)
  {
    char *const envp[] = { NULL };
  
    close(0);
    dup2(tohelper[0], 0);
    close(tohelper[0]);
    close(tohelper[1]);
    
    close(1);
    dup2(fromhelper[1], 1);
    close(fromhelper[0]);
    close(fromhelper[1]);
    
    execle(helperPath, helperPath, NULL, envp);
    _exit(1);
  }
  signal(SIGPIPE, SIG_IGN);
  
  close(tohelper[0]);
  close(fromhelper[1]);
  
  if (write(tohelper[1], &extAuth, sizeof(extAuth)) != sizeof(extAuth))
  {
    close(tohelper[1]);
    return MAHelperCommandInternalError;
  }

  //XXX do stuff to read from helper and dump on textview
  
  written= write(tohelper[1], &command, sizeof(MAHelperCommand));
  
  close(tohelper[1]);
  
  if (written != sizeof(MAHelperCommand))
    return MAHelperCommandInternalError;
  
  if (waitpid(pid, &childStatus, 0) != pid)
    return MAHelperCommandInternalError;
  
  if (!WIFEXITED(childStatus))
    return MAHelperCommandInternalError;
  
  return WEXITSTATUS(childStatus);
}



@interface MAServiceControl(Private)
- (void)updateStatus;
- (BOOL)performCommand:(MAHelperCommand)command
             withRight:(const char*)rightName;
- (void)startStopServer:(BOOL)flag;
- (void)enableServerAutoStart:(BOOL)flag;
- (char*)getStartupItem;
@end

@implementation MAServiceControl(Private)
- (void)updateStatus
{
  NSImage *image;
  NSString *status;
  NSString *blabel;
  BOOL autoStart= NO;
  
  // check if autostart is enabled
  {
    char buffer[1024];
    FILE *f;
    f= fopen("/etc/hostconfig","r");
    if (f)
    {
      while (fgets(buffer, sizeof(buffer), f))
      {
        if (strncmp(buffer,"MYSQLCOM=", sizeof("MYSQLCOM=")-1)==0)
        {
          if (strstr(buffer, "-YES-"))
          {
            autoStart= YES;
            break;
          }
        }
      }
      fclose(f);
    }
  }
  
  [startOnBootCheck setState:autoStart?NSOnState:NSOffState];

  switch ([_owner serverStatus])
  {
    case 0:
      image= [NSImage imageNamed: @"service_status_stopped.png"];
      status= @"Server is Stopped";
      blabel= @"Start Server";
      break;
    case 1:
      image= [NSImage imageNamed: @"service_status_running.png"];
      status= @"Server is Running";
      blabel= @"Stop Server";
      break;
    default:
      image= [NSImage imageNamed: @"service_status_unknown.png"];
      status= @"Cannot determine server status";
      blabel= nil;
    break;
  }
  
  [statusImage setImage: image];
  [statusLabel setStringValue: status];
  if (blabel)
  {
    [toggleButton setEnabled:YES];
    [toggleButton setTitle:blabel];
  }
  else
  {
    [toggleButton setEnabled:NO];
    [toggleButton setTitle:@"Start Server"];
  }  
}

- (char*)getStartupItem
{
  if ([[NSFileManager defaultManager] fileExistsAtPath:@"/Library/StartupItems/MySQLCOM"])
    return "/Library/StartupItems/MySQLCOM/MySQL";
  else if ([[NSFileManager defaultManager] fileExistsAtPath:@"/Library/StartupItems/MySQL"])
    return "/Library/StartupItems/MySQL/MySQL";
  else
    return NULL;
}


- (BOOL)performCommand:(MAHelperCommand)command
             withRight:(const char*)rightName
{
  AuthorizationRef authRef;
  NSString *helperPath;
  AuthorizationItem right = { rightName, 0, NULL, 0 };
  AuthorizationRights rightSet = { 1, &right };
  OSStatus status;
  
#warning fix path
  helperPath= @"/Users/akojima/guibuild/mahelper";
  
  if (AuthorizationCreate(NULL, kAuthorizationEmptyEnvironment, kAuthorizationFlagDefaults, 
                          &authRef)!=errAuthorizationSuccess)
  {
    NSLog(@"Could not create authorization reference object.");
    return NO;
  }
  
  status = AuthorizationCopyRights(authRef, &rightSet, kAuthorizationEmptyEnvironment, 
                                   kAuthorizationFlagDefaults | kAuthorizationFlagPreAuthorize
                                   | kAuthorizationFlagInteractionAllowed | kAuthorizationFlagExtendRights,
                                   NULL);
  
  if (status == errAuthorizationSuccess)
  {
    int hstatus;

    hstatus = helperPerformCommand(authRef, [helperPath UTF8String], command);
    
    AuthorizationFree(authRef, kAuthorizationFlagDestroyRights);
    
    if (hstatus == MAHelperCommandSuccess)
      return YES;
  }
  else
  {
    if (status != errAuthorizationCanceled)
    {
      // pre-auth failed
      NSLog(@"Pre-auth failed");
    }
  }
  return NO;
}


- (void)startStopServer:(BOOL)start
{
  char *path= [self getStartupItem];
  MAHelperCommand command;
  
  command.authorizedCommandId= start ? MAHelperStartMySQL : MAHelperStopMySQL;
  strcpy(command.argumentPath, path);

  [self performCommand:command
             withRight:"com.mysql.administrator.server"];
}

- (void)enableServerAutoStart:(BOOL)flag
{
  MAHelperCommand command;

  command.authorizedCommandId= MAHelperToggleAutoStart;
  command.enable= flag;
  if (![self performCommand:command
                  withRight:"com.mysql.administrator.chconfig"])
  {
    NSRunAlertPanel(@"Error", @"Could not change automatic startup configuration.", 
                    nil, nil, nil);
  }
}

@end

@implementation MAServiceControl

- (IBAction)toggleServer:(id)sender
{
  [self startStopServer:[_owner serverStatus]==0];
  [self updateStatus];
}

- (IBAction)toggleServerStartup:(id)sender
{
  [self enableServerAutoStart:[sender state]==NSOnState];
  [self updateStatus];
}

- (void)awakeFromNib
{
  [self updateStatus];
  
  if ([self getStartupItem]==NULL)
  {
    [messageText setString:@"The MySQL Startup Item was not found in /Librart/StartupItems\n"
      "You must select the \"Install Startup Item\" option during installation to enable "
      "server control from this tool."];
    [toggleButton setEnabled:NO];
    [startOnBootCheck setEnabled:NO];
  }
}

+ (NSImage*)icon
{
  return [[[NSImage alloc] initWithContentsOfFile: 
    [[NSBundle mainBundle] pathForResource:@"service_config"
                                    ofType:@"png"]] autorelease];
}

+ (NSString*)label
{
  return @"Service";
}

+ (NSString*)toolTip
{
  return @"Control MySQL Server Startup and Service.";
}

- (id)initWithOwner: (id<MAdministratorProtocol>)owner
{
  self= [super initWithNibFile: @"ServiceControl" panelOwner: owner];
  
  _defaultFrame= [[self topView] frame];
  
  return self;
}

@end
